/*
 * Copyright 2012 Steven Gribble
 *
 *  This file is part of the UW CSE333 project sequence (333proj12).
 *
 *  333proj12 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  333proj12 is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj12.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _HW1_AVLTREE_PRIV_H_
#define _HW1_AVLTREE_PRIV_H_

#include <stdint.h>      // for uint64_t

// This file defines the internal structures associated with
// our AVLTree implementation.  Customers should not include
// this file or assume anything based on its contents.  Instead,
// we have broken these out into this file so that the unit test
// code for AVLTree has access to it, allowing unit tests to
// peek inside the implementation to check pointers and fields
// for correctness.

// This struct represents an individual node within a tree.  A
// tree contains left and right pointers, a height, and a
// customer-supplied payload pointer.
typedef struct avltree_node {
  void                *payload;  // customer-supplied payload pointer
  struct avltree_node *parent;   // parent node, or NULL if root
  struct avltree_node *left;     // left child node in tree, or NULL
  struct avltree_node *right;    // right child node in tree, or NULL
  uint32_t             height;   // height of this subtree
} AVLTreeNode, *AVLTreeNodePtr;

// This struct represents the entire AVL tree.  Remember that we
// provided a struct declaration (but not definition) in AVLTree.h;
// this is the associated definition.  This struct contains metadata
// about the AVL tree.
typedef struct avltree_head {
  uint64_t          num_nodes;  //  # ndoes in the tree
  AVLTreeNodePtr    root;       // root node of the tree, or NULL if empty
} AVLTree;

// This struct represents the state of an iterator.  We expose
// the struct declaration in AVLTree.h, but not the definition, as
// above for the tree itself.
typedef struct avltree_iter {
  AVLTreePtr            tree;   // the tree we're for
  AVLTreeNodePtr        node;   // the node we print next, or NULL if done
  AVLTreeIteratorOrder  order;  // the order to traverse through
} AVLTreeIter;

// figure out the balance of a tree node, assuming the heights
// of its children are correct; is O(1)
int32_t GetBalance(AVLTreeNodePtr tn);

#endif  // _HW1_AVLTREE_PRIV_H_
