/*
 * Copyright 2012 Steven Gribble
 *
 *  This file is part of the UW CSE 333 project sequence (333proj12).
 *
 *  333proj12 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  333proj12 is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj12.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "LinkedList.h"
#include "LinkedList_priv.h"

LinkedListPtr AllocateLinkedList(void) {
  // allocate the linked list record
  LinkedListPtr ll = (LinkedListPtr) malloc(sizeof(LinkedList));
  if (ll == NULL) {
    // out of memory
    return (LinkedListPtr) NULL;
  }

  // initialize the newly allocated record structure
  // STEP 1.

  // return our newly minted linked list
  return ll;
}

void FreeLinkedList(LinkedListPtr list,
                    LLPayloadFreeFnPtr payload_free_function) {
  // defensive programming: check arguments for sanity.
  assert(list != NULL);
  assert(payload_free_function != NULL);

  // sweep through the list and free all of the nodes' payloads as
  // well as the nodes themselves
  while (list->head != NULL) {
    // STEP 2.

  }

  // free the list record
  free(list);
}

uint64_t NumElementsInLinkedList(LinkedListPtr list) {
  // defensive programming: check argument for safety.
  assert(list != NULL);
  return list->num_elements;
}

int PushLinkedList(LinkedListPtr list, void *payload) {
  // defensive programming: check argument for safety; the
  // user-supplied argument can be anything, of course.
  assert(list != NULL);

  // allocate space for the new node.
  // STEP 3.
  LinkedListNodePtr ln = NULL;  // you might need to modify this line too.

  // set the payload
  ln->payload = payload;

  if (list->num_elements == 0) {
    // degenerate case; list is currently empty
    assert(list->head == NULL);  // debugging aid
    assert(list->tail == NULL);  // debugging aid
    ln->next = ln->prev = NULL;
    list->head = list->tail = ln;
    list->num_elements = 1U;
    return 1;
  }

  // typical case; list has >=1 elements
  assert(list->head != NULL);  // debugging aid
  assert(list->tail != NULL);  // debugging aid
  // STEP 4.  Add the element to the head of the non-empty list.

  // return success
  return 1;
}

int PopLinkedList(LinkedListPtr list, void **payload_ptr) {
  // defensive programming.
  assert(payload_ptr != NULL);
  assert(list != NULL);

  // if there is nothing in the list, fail.
  if (list->num_elements == 0) {
    return 0;
  }

  // There are two cases to consider: (a) a list with a single element
  // in it and (b) a list with >=2 elements in it.
  // STEP 5.


  // depending on what you did in step 5, you might need
  // to change this return to some other value.
  return 1;
}

int AppendLinkedList(LinkedListPtr list, void *payload) {
  // defensive programming: check argument for safety; the
  // user-supplied argument can be anything, of course.
  assert(list != NULL);

  // degenerate case; list is currently empty
  if (list->num_elements == 0)
    return PushLinkedList(list, payload);

  // typical case; list has >=1 elements.
  // allocate space for the new node, and splice it in.
  // STEP 6.


  // return success
  return 1;
}

int SliceLinkedList(LinkedListPtr list, void **payload_ptr) {
  // defensive programming.
  assert(payload_ptr != NULL);
  assert(list != NULL);

  // if there is nothing in the list, fail.
  if (list->num_elements == 0) {
    return 0;
  }

  // There are two cases to consider: (a) a list with a single element
  // in it and (b) a list with >=2 elements in it.
  // STEP 7.

  // depending on what you did above in step 7, you might need to
  // change this return to do something else.
  return 0;
}

void SortLinkedList(LinkedListPtr list, unsigned int ascending,
                    LLPayloadComparatorFnPtr comparator_function) {
  assert(list != NULL);  // defensive programming
  if (list->num_elements < 2) {
    // no sorting needed
    return;
  }

  // we'll implement bubblesort! nice and easy, and nice and slow :)
  int swapped;
  do {
    LinkedListNodePtr curnode;

    swapped = 0;
    curnode = list->head;
    while (curnode->next != NULL) {
      int compare_result = comparator_function(curnode->payload,
                                               curnode->next->payload);
      if (ascending) {
        compare_result *= -1;
      }
      if (compare_result < 0) {
        // bubble-swap payloads
        void *tmp;
        tmp = curnode->payload;
        curnode->payload = curnode->next->payload;
        curnode->next->payload = tmp;
        swapped = 1;
      }
      curnode = curnode->next;
    }
  } while (swapped);
}

LLIterPtr LLMakeIterator(LinkedListPtr list, int pos) {
  // defensive programming
  assert(list != NULL);
  assert((pos == 0) || (pos == 1));

  // if the list is empty, return failure.
  if (NumElementsInLinkedList(list) == 0U)
    return NULL;

  // OK, let's manufacture an iterator.
  LLIterPtr li = (LLIterPtr) malloc(sizeof(LLIter));
  if (li == NULL) {
    // out of memory!
    return NULL;
  }

  // set up the iterator.
  li->list = list;
  if (pos == 0) {
    li->node = list->head;
  } else {
    li->node = list->tail;
  }

  // return the new iterator
  return li;
}

void LLIteratorFree(LLIterPtr iter) {
  // defensive programming
  assert(iter != NULL);
  free(iter);
}

int LLIteratorHasNext(LLIterPtr iter) {
  // defensive programming
  assert(iter != NULL);
  assert(iter->list != NULL);
  assert(iter->node != NULL);

  // Is there another node beyond the iterator?
  if (iter->node->next == NULL)
    return 0;  // no

  return 1;  // yes
}

int LLIteratorNext(LLIterPtr iter) {
  // defensive programming
  assert(iter != NULL);
  assert(iter->list != NULL);
  assert(iter->node != NULL);

  // if there is another node beyond the iterator, advance to it.
  // STEP 8.


  // nope, there isn't, so return failure.
  return 0;
}

int LLIteratorHasPrev(LLIterPtr iter) {
  // defensive programming
  assert(iter != NULL);
  assert(iter->list != NULL);
  assert(iter->node != NULL);

  // Is there another node beyond the iterator?
  if (iter->node->prev == NULL)
    return 0;  // no

  return 1;  // yes
}

int LLIteratorPrev(LLIterPtr iter) {
  // defensive programming
  assert(iter != NULL);
  assert(iter->list != NULL);
  assert(iter->node != NULL);

  // if there is another node beyond the iterator, advance to it.
  if (iter->node->prev != NULL) {
    iter->node = iter->node->prev;
    return 1;
  }

  // nope, so return failure.
  return 0;
}

void LLIteratorGetPayload(LLIterPtr iter, void **payload) {
  // defensive programming
  assert(iter != NULL);
  assert(iter->list != NULL);
  assert(iter->node != NULL);

  // set the return parameter.
  *payload = iter->node->payload;
}

int LLIteratorDelete(LLIterPtr iter,
                     LLPayloadFreeFnPtr payload_free_function) {
  // defensive programming
  assert(iter != NULL);
  assert(iter->list != NULL);
  assert(iter->node != NULL);

  // degenerate case: the list becomes empty after deleting.
  if (NumElementsInLinkedList(iter->list) == 1U) {
    void *payload;
    PopLinkedList(iter->list, &payload);
    payload_free_function(payload);
    iter->node = NULL;  // defensive maneuver
    return 0;
  }

  // degenerate case: iter points at head
  if (iter->node == iter->list->head) {
    // use our own poplinkedlist
    void *payload;
    PopLinkedList(iter->list, &payload);
    payload_free_function(payload);
    iter->node = iter->list->head;
    return 1;
  }

  // degenerate case: iter points at tail
  if (iter->node == iter->list->tail) {
    // use our own slicelinkedlist
    void *payload;
    SliceLinkedList(iter->list, &payload);
    payload_free_function(payload);
    iter->node = iter->list->tail;
    return 1;
  }

  // fully general case; we have to splice.
  // STEP 9.


  return 1;
}

int LLIteratorInsertBefore(LLIterPtr iter, void *payload) {
  // defensive programming
  assert(iter != NULL);
  assert(iter->list != NULL);
  assert(iter->node != NULL);

  // If the cursor is pointing at the head, use our
  // PushLinkedList function.
  if (iter->node == iter->list->head) {
    return PushLinkedList(iter->list, payload);
  }

  // General case: we have to do some splicing.
  LinkedListNodePtr newnode =
    (LinkedListNodePtr) malloc(sizeof(LinkedListNode));
  if (newnode == NULL)
    return 0;  // out of memory

  newnode->payload = payload;
  newnode->next = iter->node;
  newnode->prev = iter->node->prev;
  newnode->prev->next = newnode;
  newnode->next->prev = newnode;
  iter->list->num_elements += 1;
  return 1;
}
